var xmlrpc = require('xmlrpc');
var request = require('request');
var neoxmlinutil = require('./neoxmlinutil');
var http = require('http');
var https = require('https');
var fs = require('fs');
var xml2js = new require('xml2js');
var xml2jsParser = require('xml2js').parseString;
var xml2jsBuilder = new xml2js.Builder();
var url = require("url");
var strftime = require('strftime');

var options;
var config;
var logger;

var server;

var delayBeforeRetryTopUpReport = 10 * 1000;
var maxTopUpReportRetry = 5;

function start(_options) {
    options = _options;

    if (options.config) {
        config = options.config;
    }
    else {
        config = require("./config.json");
    }

    if (options.logger) {
        logger = options.logger;
    }
    else {
        logger = console;
    }

    createResponseServer();

    createDiyHttpXmlRpcServer();
    //createXmlRpcServer()
    //createExpressXmlRpcServer();

    try {
        http.globalAgent.maxSockets = Infinity;
    }
    catch(e) {
        logger.warn('Error setting http.globalAgent.maxSockets = Infinity', {err: e});
    }

}

function getXmlRpcParam(values) {
    try {

        var count = values.length
        var result = {};
        for (var i = 0; i < count; i++) {
            var value = values[i];

            var keys = Object.keys(value.value[0]);
            var firstKey = keys[0];
            result[value.name[0]] = value.value[0][firstKey][0];
        }

        return result;

    }
    catch(err) {
        return null;
    }
}

function createDiyHttpXmlRpcServer() {
    var serverOptions = {
        key: fs.readFileSync('./server.key'),
        cert: fs.readFileSync('./server.crt')
    }

    var httpServer = https.createServer(serverOptions, function(req, res) {
        var remoteAddress = req.connection.remoteAddress.replace(/^::ffff:/, '');

        logger.verbose("Incoming connection from " + remoteAddress);

        var body = "";

        req.on('data', function (chunk) {
            body += chunk;
        });

        req.on('end', function () {

            xml2jsParser(body, function(err, message) {

                if (err) {
                    logger.warn('Got XMLRPC invalid XML request body from ' + remoteAddress, {body: body, err: err});
                    res.end('Invalid XMLRPC request');
                    return;
                }

                var method;
                var _params;

                try {
                    method = message.methodCall.methodName[0];
                    _params = message.methodCall.params[0].param[0].value[0].struct[0].member;
                }
                catch(errSelectMethod) {
                    logger.warn('Failed to get method and params on request from ' + remoteAddress);
                    res.end('Invalid XMLRPC message')
                    return;
                }

                var params = getXmlRpcParam(_params);
                logger.info('Got XMLRPC request from ' + remoteAddress, {method: method, params: params});

                if (method == 'topUpRequest') {

                    sendTopUpRequestToMaster(params, remoteAddress, function(forwardError) {
                        if (forwardError) {
                            immediateReply(params, res, '40', forwardError.toString());
                        } else {
                            immediateReply(params, res, '68');
                        }
                    });

                }
                else if (method == 'topUpInquiry') {

                    sendTopUpInquryToMaster(params, remoteAddress, res);

                }
                else {
                    res.end('Invalid method');
                }


            })
        });

    });

    httpServer.listen(config.server_options.port, function() {
        logger.info('HTTP XMLRPC listen on port ' + config.server_options.port);
    });
}

function composeXmlRpcResponse(param) {

    var values = [];
    var keys = Object.keys(param);
    var keysCount = keys.length;

    for (var i = 0; i < keysCount; i++) {
        var key = keys[i];
        var value = {
            name: key,
            value: {
                string: param[key]
            }
        }
        values.push(value);
    }

    var data = {
        methodResponse: {
            params: {
                param: {
                    value: {
                        struct: {
                            member: values
                        }
                    }
                }
            }
        }
    }

    //logger.info(JSON.stringify(data));

    return xml2jsBuilder.buildObject(data);
}

function immediateReply(param, requestResponse, responseCode, message) {

    if (!responseCode) {
        responseCode = '68';
    }

    if (!message) {
        message = "Sementara tidak dapat diproses.";
    }

    if (responseCode == '68') {
        message = 'ISI '
            + param.NOM.toUpperCase()
            + ' KE '
            + param.NOHP
            + ', Transaksi anda sedang diproses'
    }

    var responseData = {
        'RESPONSECODE': responseCode,
        'REQUESTID': param.REQUESTID,
        'MESSAGE': message,
        'TRANSACTIONID': '0',
    }


    var responseBody = composeXmlRpcResponse(responseData)
    logger.info("Sending immediateReply response", {responseData: responseData});

    if (config.dump_xml_to_log) {
        logger.info("immediateReply XML BODY", {xmlbody: responseBody});
    }

    requestResponse.writeHead(200, {'Content-Type': 'text/xml'});
    requestResponse.end(responseBody);
}

function composeMessage(params, remoteAddress) {
    try {
        var nom = params.NOM.replace(/\./g, '').trim();
        var destination = params.NOHP.replace(/\./g, '').trim();
        var pin = params.PIN.replace(/\./g, '').trim();
        var requestId = params.REQUESTID.replace(/\./g, '').trim();

        return 'MI.' + nom + '."' + destination + '".' + pin + '.' + requestId + '.NOTRUST."' + remoteAddress + '"';
    }
    catch(err) {
        return;
    }


}

function inquiryReply(responseCode, message, responseToClient) {
    var responseData = {
        'RESPONSECODE': responseCode,
        'MESSAGE': message
    }

    var responseBody = composeXmlRpcResponse(responseData)
    logger.info("Sending inquiry response", {responseData: responseData});

    responseToClient.writeHead(200, {'Content-Type': 'text/xml'});
    responseToClient.end(responseBody);
}

function parseTopUpInquiryResponse(raw) {
    var tokens = raw.split('$');
    if (tokens.length < 2) {
        return {
            responseCode: '68',
            message: 'Gagal parsing hasil inquiry'
        }
    }

    var responseCode = tokens[0];
    var message = '';

    for (var i = 1; i < tokens.length; i++) {
        message += '$' + tokens[i];
    }

    return {
        responseCode: responseCode,
        message: message.replace('$', '')
    }
}

function sendTopUpInquryToMaster(params, remoteAddress, responseToClient) {
    if (!config.inquiry_url) {
        inquiryReply('68', 'Inquiry server tidak terdaftar');
        return;
    }

    var requestOpts = {
		url: config.inquiry_url,
		qs: {
            REQUESTID: params.REQUESTID,
			MSISDN: params.MSISDN,
			PIN: params.PIN,
			NO_HP: params.NOHP || params.NO_HP,
            NOHP: params.NOHP || params.NO_HP,
            TRUST: 'NO_TRUST',
            IP_ADDR: remoteAddress
		}
	};

    request(requestOpts, function(requestError, response, body) {

        if (requestError) {
            logger.warn('Error requesting to inquiry server: ' + requestError);
            inquiryReply('68', 'Gagal konek ke inquiry server', responseToClient);
            return;
        }

        logger.info('Got response from inquiry server', {body: body, requestOpts: requestOpts});
        var params = parseTopUpInquiryResponse(body);
        inquiryReply(params.responseCode, params.message, responseToClient);
    })
}

function sendTopUpRequestToMaster(param, remoteAddress, callback) {

    /*
    var smscidSuffix = '99999999999999' + String(Math.round(Math.random() * 99999999999999));
    var smscid = 'XML1' + smscidSuffix.slice(-13);
    */

    //var smscid = 'XML' + '12345' + strftime('%H%M%S%L');
    var smscid = config.smscid;// + Math.round(Math.random() * 9999999999999);

    var message = composeMessage(param, remoteAddress);
    if (!message) {
        logger.warn('Invalid message for sendToMaster', {param: param});
        return;
    }

    var requestOpts = {
		url: config.master_url,
		qs: {
			PhoneNumber: param.MSISDN,
			'Text': message,
			Res_Port: config.res_port,
			Smscid: smscid
		}
	};

    if (config.res_ip) {
        requestOpts.qs.ip_addr = config.res_ip;
    }

    logger.info('Forward request to master', {requestOpts: requestOpts});

    request(requestOpts, function(requestError, response, body) {

        if (requestError) {
            logger.warn('Error requesting to master: ' + requestError);
        } else {
            logger.info('Got response from master', {body: body, requestOpts: requestOpts});
        }

        callback(requestError);
    })

}

function sendReply(response) {
    /*
    var requestId = neoxmlinutil.getRequestIdFromResponseMessage(response.text);

    if (!requestId) {
        logger.warn('No request id found, skipping');
        return;
    }
    */

    var params = {
        REQUESTID: response.requestid,
        MSISDN: response.PhoneNumber,
        RESPONSECODE: response.resp_code,
        MESSAGE: response.text,
    }

    if (response.new_sn) {
        params.SN = response.new_sn.replace(/^SN=/, '');
    }

    if (response.new_ending_balance) {
        params.BALANCE = response.new_ending_balance;
    }

    if (response.new_price) {
        params.PRICE = response.new_price;
    }

    logger.info('sendReply', {params: params});

    neoxmlinutil.getReverseUrl(response.PhoneNumber, function(err, reverseUrls) {
        if (err) {
            logger.warn('Fail to get reverse urls, skipping');
            return;
        }

        if (reverseUrls.length <= 0) {
            logger.warn('No reverse urls found, skipping');
            return;
        }

        sendTopUpReport(reverseUrls, params, 0, 4);
    });
}

function errorOnTopUpReport(topUpReportError, debugInfo) {
    var hasError = false;

    if (!topUpReportError) {
        return false;
    }

    try {
        if (topUpReportError.body) {
            logger.verbose('errorOnTopUpReport response body: ' + topUpReportError.body, {req: debugInfo});
        }

        if (topUpReportError.res && topUpReportError.res.statusCode && topUpReportError.res.statusCode != '200') {
            logger.verbose('Partner response with status code:' + topUpReportError.res.statusCode, {req: debugInfo});

            if (topUpReportError.res.statusCode != '200') {
                logger.warn('Going to resend topUpReport because status code is not 200 (' + topUpReportError.res.statusCode + ')', {req: debugInfo});
                return true;
            }
        }
    } catch(e) {}

    try {
        if (topUpReportError.toString().indexOf('Invalid XML-RPC message') < 0) {
            logger.verbose('errorOnTopUpReport', {error: topUpReportError.toString()});
            return true;
        }
    }
    catch(e) {}

    return false;

}

function sendTopUpReport(reverseUrls, params, urlIdx, retry) {
    if (retry === null || retry === undefined) {
        retry = maxTopUpReportRetry;
    }

    if (urlIdx === null || urlIdx === undefined) {
        urlIdx = 0;
    }

    if (urlIdx >= reverseUrls.length) {
        logger.warn('No other reverse urls for partner available', {reverseUrls: reverseUrls, params: params});

        if (retry > 0) {
            logger.warn('Retrying to send topUpReport to partner first url', {remaining_retry: retry-1, reverseUrls: reverseUrls, params: params});
            setTimeout(
                sendTopUpReport,
                delayBeforeRetryTopUpReport,
                reverseUrls,
                params,
                0,
                --retry
            )
            return;
        }

        logger.warn('topUpReport retry exceed', {reverseUrls: reverseUrls, params: params});
        return;
    }

    var partnerUrl = url.parse(reverseUrls[urlIdx]);

    var clientOptions = {
        host: partnerUrl.hostname
        , port: partnerUrl.port
        , path: partnerUrl.pathname
    };

    //logger.verbose('Preparing to send topUpReport', {clientOptions: clientOptions});

    var client;
    if (partnerUrl.protocol == 'https:') {
        client = xmlrpc.createSecureClient(clientOptions);
    } else {
        client = xmlrpc.createClient(clientOptions);
    }

    var methodName = 'topUpReport';
    logger.info('Requesting topUpReport', {protocol: partnerUrl.protocol, clientOptions: clientOptions, params: params});

    client.methodCall(methodName, [ params ], function (topUpReportError, value) {

        if (errorOnTopUpReport(topUpReportError, {params: params, clientOptions: clientOptions})) {

            logger.warn('Error sending topUpReport retrying another url (if available): ' + topUpReportError,
                {
                    error: topUpReportError.toString(),
                    responseBody: topUpReportError.body,
                    params: params,
                    clientOptions: clientOptions,
                }
            );
            sendTopUpReport(reverseUrls, params, ++urlIdx, retry);
            return;

        }

        if (config.dump_xml_to_log) {
            logger.verbose('sendTopUpReport XML BODY', {xmlbody: value});
        }

        logger.verbose("topUpReport ACK", {params: params, clientOptions: clientOptions});

    });
}

function createResponseServer() {
    var httpServer = http.createServer(function(req, res) {

        res.end();

        var parsed_url = url.parse(req.url, true, true);
        logger.info("Hit on response server", {qs: parsed_url.query});

        sendReply(parsed_url.query);

    });

    httpServer.listen(config.res_port, function() {
        logger.info('HTTP Response server listen on port ' + config.res_port);
    });
}

exports.start = start;
